package ru.landgrafhomyak.bgtu.db0.storage.sqlite_jdbc

import java.sql.Connection
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.sql.Types
import kotlin.contracts.ExperimentalContracts
import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import ru.landgrafhomyak.bgtu.db0.storage.api.ModeratorPermissions
import ru.landgrafhomyak.bgtu.db0.storage.api.QueueSet
import ru.landgrafhomyak.bgtu.db0.storage.api.Storage

class JdbcSqliteStorage(
	private val connection: Connection
) : Storage {

	init {
		this.connection.autoCommit = false
		this.connection.executeStatement("ROLLBACK TRANSACTION")
		this.connection.executeStatement("PRAGMA foreign_keys=TRUE")
	}


	@Suppress("FunctionName")
	private fun _nothingInserted(): Nothing = throw RuntimeException("Nothing inserted, but must be")

	private fun PreparedStatement.executeQueryReturningRowId(): Long = this@executeQueryReturningRowId.executeQuery { rs ->
		if (!rs.next())
			this@JdbcSqliteStorage._nothingInserted()
		return@executeQueryReturningRowId rs.getLong(1)
	}

	override fun allocQueue(chat: Long, message: Long, setName: String, queueNames: List<String>): Unit = this.connection.transaction { conn ->
		val chatDbId: Long
		conn.prepareStatement(
			"""
				INSERT INTO chats("telegram::chat_id")
				VALUES (?)
				ON CONFLICT DO UPDATE SET "telegram::chat_id" = EXCLUDED."telegram::chat_id"
				RETURNING _internal_id
			"""
		) { ps ->
			ps.setLong(1, chat)
			chatDbId = ps.executeQueryReturningRowId()
		}

		val queueSetDbId: Long
		this.connection.prepareStatement(
			"""
				INSERT INTO queue_sets(chat, "telegram::message_id", name)
				VALUES (?, ?, ?)
				RETURNING _internal_id
			"""
		) { ps ->
			ps.setLong(1, chatDbId)
			ps.setLong(2, message)
			ps.setString(3, setName)
			queueSetDbId = ps.executeQueryReturningRowId()
		}
		this.connection.prepareStatement(
			"""
				INSERT INTO queues(queue_set, name, ordinal)
				VALUES (?, ?, ?)
			"""
		) { ps ->
			ps.setLong(1, queueSetDbId)

			queueNames.forEachIndexed { i, n ->
				ps.setString(2, n)
				ps.setInt(3, i)
				ps.addBatch()
			}
			ps.executeBatch()
		}
	}

	@Suppress("FunctionName")
	private fun ResultSet.__formatQueue(ordinalColumnIndex: Int, tgUidColumnIndex: Int, displayNameColumnIndex: Int): List<UserStruct?> {
		val users = ArrayList<UserStruct?>()
		while (this@__formatQueue.next()) {
			val userOrdinal = this@__formatQueue.getInt(ordinalColumnIndex)
			val user = UserStruct(
				telegramId = this@__formatQueue.getLong(tgUidColumnIndex),
				displayName = this@__formatQueue.getString(displayNameColumnIndex)
			)
			while (users.size < userOrdinal)
				users.add(null)

			users.add(user)
		}

		return users
	}

	@Suppress("FunctionName")
	private fun _formatQueueSet(conn: Connection, qs: FoundQueue): QueueSet {
		val queuesMeta = conn.prepareStatement(
			"""
				SELECT _internal_id, name
				FROM queues
				WHERE queue_set = ?
				ORDER BY ordinal ASC
			"""
		) { ps ->
			ps.setLong(1, qs.dbId)
			return@prepareStatement ps.executeQuery()
				.mapToThenClose(LinkedHashSet()) { rs -> Pair(rs.getLong(1), rs.getString(2)) }
		}


		val queues = ArrayList<QueueStruct>()
		conn.prepareStatement(
			"""
				SELECT "user->queue".ordinal, users."telegram::user_id", users.display_name
				FROM "user->queue"
					INNER JOIN users ON users._internal_id = "user->queue".user
				WHERE "user->queue".${if (queuesMeta.isEmpty()) "queue_set" else "queue"} = ?
				ORDER BY "user->queue".ordinal ASC
			"""
		) { ps ->
			if (queuesMeta.isEmpty()) {
				ps.setLong(1, qs.dbId)
				ps.executeQuery { rs -> queues.add(QueueStruct(null, rs.__formatQueue(1, 2, 3))) }
			} else {
				for ((queueDbId, queueName) in queuesMeta) {
					ps.setLong(1, queueDbId)
					ps.executeQuery { rs -> queues.add(QueueStruct(queueName, rs.__formatQueue(1, 2, 3))) }
				}
			}
		}

		return QueueSetStruct(
			name = qs.name,
			queues = queues,
			isOpen = qs.isOpen
		)
	}

	private class FoundQueue(
		val dbId: Long,
		val isOpen: Boolean,
		val name: String
	) {
		fun withIsOpen(newValue: Boolean) = FoundQueue(
			dbId = this.dbId,
			isOpen = newValue,
			name = this.name
		)
	}

	@Suppress("FunctionName", "IMPLICIT_NOTHING_TYPE_ARGUMENT_AGAINST_NOT_NOTHING_EXPECTED_TYPE")
	private fun _findQueueSet(conn: Connection, chat: Long, message: Long): FoundQueue? = conn.prepareStatement(
		"""
			SELECT queue_sets._internal_id, queue_sets.is_open, queue_sets.name
			FROM queue_sets 
				INNER JOIN chats on chats._internal_id = queue_sets.chat
			WHERE chats."telegram::chat_id" = ? AND queue_sets."telegram::message_id" = ?
		"""
	) { ps ->
		ps.setLong(1, chat)
		ps.setLong(2, message)
		ps.executeQuery { rs ->
			if (rs.next())
				return@_findQueueSet FoundQueue(rs.getLong(1), rs.getBoolean(2), rs.getString(3))
			else
				return@_findQueueSet null
		}
	}


	@Suppress("FunctionName")
	private fun _closeOrOpenQueue(chat: Long, message: Long, newIsOpen: Boolean): Storage.Result_QueueCloseOrOpen = this.connection.transaction { conn ->
		val qs = this._findQueueSet(conn, chat, message) ?: return Storage.Result_QueueCloseOrOpen.QUEUE_NOT_FOUND

		if (qs.isOpen == newIsOpen)
			return Storage.Result_QueueCloseOrOpen.ALREADY_IN_STATE

		conn.prepareStatement(
			"""
				UPDATE queue_sets 
				SET is_open = ?
				WHERE _internal_id = ?
			"""
		) { ps ->
			ps.setBoolean(1, newIsOpen)
			ps.setLong(2, qs.dbId)
			ps.executeUpdate()
		}

		return Storage.Result_QueueCloseOrOpen.SUCCESSFUL(this._formatQueueSet(conn, qs.withIsOpen(newIsOpen)))
	}

	override fun closeQueue(chat: Long, message: Long): Storage.Result_QueueCloseOrOpen = this._closeOrOpenQueue(chat, message, false)

	override fun reopenQueue(chat: Long, message: Long): Storage.Result_QueueCloseOrOpen = this._closeOrOpenQueue(chat, message, true)


	@Suppress("FunctionName")
	private inline fun ResultSet._getFirstFreeOrdinal(ordinalColumnIndex: Int = 1, extraActions: (ResultSet) -> Unit = {}): Int {
		var last = -1
		while (this@_getFirstFreeOrdinal.next()) {
			val current = this@_getFirstFreeOrdinal.getInt(ordinalColumnIndex)
			if (last + 1 < current)
				return last + 1
			last = current
			extraActions(this@_getFirstFreeOrdinal)
		}
		return last + 1
	}

	@OptIn(ExperimentalContracts::class)
	@Suppress("FunctionName")
	private inline fun <R> Connection._prepareStatementWithQueueDbId(
		queueIdParamIndex: Int,
		queueSetDbId: Long,
		queueDbId: Long?,
		queryBuilder: (isNull: Boolean) -> String,
		extra: (PreparedStatement) -> R
	): R {
		contract {
			callsInPlace(queryBuilder, InvocationKind.EXACTLY_ONCE)
			callsInPlace(extra, InvocationKind.EXACTLY_ONCE)
		}

		if (queueDbId == null) {
			return this@_prepareStatementWithQueueDbId.prepareStatement(queryBuilder(true)) { ps ->
				ps.setLong(queueIdParamIndex, queueSetDbId)
				return@prepareStatement extra(ps)
			}
		} else {
			return this@_prepareStatementWithQueueDbId.prepareStatement(queryBuilder(false)) { ps ->
				ps.setLong(queueIdParamIndex, queueDbId)
				return@prepareStatement extra(ps)
			}
		}
	}

	override fun addUserToQueue(chat: Long, message: Long, user: Long, userDisplayName: String): Storage.Result_AddOrRemoveUser = this.connection.transaction { conn ->
		val qs = this._findQueueSet(conn, chat, message) ?: return Storage.Result_AddOrRemoveUser.QUEUE_NOT_FOUND
		if (!qs.isOpen) return Storage.Result_AddOrRemoveUser.QUEUE_IS_FINAL


		val userDbId: Long
		conn.prepareStatement(
			"""
				INSERT INTO users("telegram::user_id", display_name)
				VALUES (?, ?)
				ON CONFLICT("telegram::user_id") DO UPDATE SET display_name = EXCLUDED.display_name
				RETURNING _internal_id
			"""
		) { ps ->
			ps.setLong(1, user)
			ps.setString(2, userDisplayName)
			userDbId = ps.executeQueryReturningRowId()
		}


		conn.prepareStatement(
			"""
				SELECT TRUE
				FROM "user->queue" 
				WHERE queue_set = ? AND user = ? 
			"""
		) { ps ->
			ps.setLong(1, qs.dbId)
			ps.setLong(2, userDbId)
			ps.executeQuery { rs ->
				if (rs.next())
					return Storage.Result_AddOrRemoveUser.ALREADY_IN_STATE
			}
		}


		val firstQueueDbId: Long?
		conn.prepareStatement(
			"""
				SELECT _internal_id 
				FROM queues
				WHERE queue_set = ?
				ORDER BY ordinal
				LIMIT 1
			"""
		) { ps ->
			ps.setLong(1, qs.dbId)
			firstQueueDbId = ps.executeQuery { rs -> if (!rs.next()) return@executeQuery null else return@executeQuery rs.getLong(1) }
		}

		val userOrdinal: Int
		conn._prepareStatementWithQueueDbId(
			1, qs.dbId, firstQueueDbId,
			{ isNull ->
				"""
					SELECT ordinal
					FROM "user->queue"
					WHERE ${if (isNull) "queue_set" else "queue"} = ?
					ORDER BY ordinal ASC	
				"""
			}
		) { ps ->
			userOrdinal = ps.executeQuery { rs -> rs._getFirstFreeOrdinal() }
		}


		conn.prepareStatement(
			"""
				INSERT INTO "user->queue"(user, queue_set, queue, ordinal) 
				VALUES (?, ?, ?, ?) 
			"""
		) { ps ->
			ps.setLong(1, userDbId)
			ps.setLong(2, qs.dbId)
			if (firstQueueDbId == null)
				ps.setNull(3, Types.INTEGER)
			else
				ps.setLong(3, firstQueueDbId)
			ps.setInt(4, userOrdinal)
			ps.executeUpdate()
		}

		return Storage.Result_AddOrRemoveUser.SUCCESSFUL(this._formatQueueSet(conn, qs))
	}

	override fun removeUserFromQueue(chat: Long, message: Long, user: Long): Storage.Result_AddOrRemoveUser = this.connection.transaction { conn ->
		val qs = this._findQueueSet(conn, chat, message) ?: return Storage.Result_AddOrRemoveUser.QUEUE_NOT_FOUND
		if (!qs.isOpen) return Storage.Result_AddOrRemoveUser.QUEUE_IS_FINAL


		conn.prepareStatement(
			"""
				DELETE FROM "user->queue"
				WHERE queue_set = ? AND user = (SELECT _internal_id FROM users WHERE "telegram::user_id" = ?)
				RETURNING TRUE
			"""
		) { ps ->
			ps.setLong(1, qs.dbId)
			ps.executeQuery { rs ->
				if (!rs.next())
					return Storage.Result_AddOrRemoveUser.ALREADY_IN_STATE
			}
		}


		return Storage.Result_AddOrRemoveUser.SUCCESSFUL(this._formatQueueSet(conn, qs))
	}

	@Suppress("FunctionName")
	private fun _moveUserToAnotherQueue(chat: Long, message: Long, user: Long, moveToNext: Boolean): Storage.Result_MoveToNextOrPrevQueue = this.connection.transaction { conn ->
		val qs = this._findQueueSet(conn, chat, message) ?: return Storage.Result_MoveToNextOrPrevQueue.QUEUE_NOT_FOUND
		if (!qs.isOpen) return Storage.Result_MoveToNextOrPrevQueue.QUEUE_IS_FINAL

		val userDbId: Long
		val queueDbId: Long
		conn.prepareStatement(
			"""
				SELECT users._internal_id, "user->queue".queue
				FROM "user->queue"
					INNER JOIN users ON users._internal_id = "user->queue".user
				WHERE users."telegram::user_id" = ? AND "user->queue".queue_set = ?
			"""
		) { ps ->
			ps.setLong(1, user)
			ps.setLong(2, qs.dbId)
			ps.executeQuery { rs ->
				if (!rs.next())
					return@_moveUserToAnotherQueue Storage.Result_MoveToNextOrPrevQueue.NOT_IN_QUEUE

				userDbId = rs.getLong(1)
				queueDbId = rs.getLong(2)
				if (rs.wasNull()) {
					return@_moveUserToAnotherQueue Storage.Result_MoveToNextOrPrevQueue.ON_EDGE
				}
			}
		}

		val newQueueDbId: Long
		conn.prepareStatement(
			"""
				SELECT _internal_id
				FROM queues
				WHERE queue_set = ?
				  AND ordinal ${if (moveToNext) '>' else '<'} (SELECT current_queue.ordinal 
								                               FROM queues AS current_queue
								                               WHERE current_queue._internal_id = ?)
			    ORDER BY ordinal ${if (moveToNext) "ASC" else "DESC"}
				LIMIT 1
			"""
		) { ps ->
			ps.setLong(1, qs.dbId)
			ps.setLong(2, queueDbId)
			ps.executeQuery { rs ->
				if (!rs.next())
					return@_moveUserToAnotherQueue Storage.Result_MoveToNextOrPrevQueue.ON_EDGE
				newQueueDbId = rs.getLong(1)
			}
		}

		val newOrdinal: Int
		conn.prepareStatement(
			"""
					SELECT ordinal
					FROM "user->queue"
					WHERE queue = ?
					ORDER BY ordinal ASC
				"""
		) { ps ->
			ps.setLong(1, newQueueDbId)
			newOrdinal = ps.executeQuery { rs -> rs._getFirstFreeOrdinal() }
		}

		conn.prepareStatement(
			"""
				UPDATE "user->queue"
				SET queue = ?, ordinal = ?
				WHERE user = ? AND queue_set = ?
			"""
		) { ps ->
			ps.setLong(1, newQueueDbId)
			ps.setInt(2, newOrdinal)
			ps.setLong(3, userDbId)
			ps.setLong(4, qs.dbId)
			ps.executeUpdate()
		}

		return Storage.Result_MoveToNextOrPrevQueue.SUCCESSFUL(this._formatQueueSet(conn, qs))
	}

	override fun moveUserToNextQueue(chat: Long, message: Long, user: Long): Storage.Result_MoveToNextOrPrevQueue =
		this._moveUserToAnotherQueue(chat, message, user, true)

	override fun moveUserToPrevQueue(chat: Long, message: Long, user: Long): Storage.Result_MoveToNextOrPrevQueue =
		this._moveUserToAnotherQueue(chat, message, user, false)

	@Suppress("FunctionName")
	private fun _moveUserInQueue(chat: Long, message: Long, user: Long, moveToStart: Boolean): Storage.Result_MoveToStartOrEnd = this.connection.transaction { conn ->
		val qs = this._findQueueSet(conn, chat, message) ?: return Storage.Result_MoveToStartOrEnd.QUEUE_NOT_FOUND
		if (!qs.isOpen) return Storage.Result_MoveToStartOrEnd.QUEUE_IS_FINAL

		val userDbId: Long
		val queueDbId: Long?
		val currentOrdinal: Int
		conn.prepareStatement(
			"""
				SELECT users._internal_id, "user->queue".queue, "user->queue".ordinal
				FROM "user->queue"
					INNER JOIN users ON users._internal_id = "user->queue".user
				WHERE users."telegram::user_id" = ? AND "user->queue".queue_set = ?
			"""
		) { ps ->
			ps.setLong(1, user)
			ps.setLong(2, qs.dbId)
			ps.executeQuery { rs ->
				if (!rs.next())
					return@_moveUserInQueue Storage.Result_MoveToStartOrEnd.NOT_IN_QUEUE

				userDbId = rs.getLong(1)
				queueDbId = rs.getLong(2).takeUnless { rs.wasNull() }
				currentOrdinal = rs.getInt(3)
			}
		}

		val newOrdinal: Int
		conn._prepareStatementWithQueueDbId(
			1, qs.dbId, queueDbId,
			{ isNull ->
				"""
					SELECT ordinal
					FROM "user->queue"
					WHERE ${if (isNull) "queue_set" else "queue"} = ?
					  AND ordinal ${if (moveToStart) '<' else '>'} ?
					ORDER BY ordinal ${if (moveToStart) "DESC" else "ASC"}
				"""
			}
		) { ps ->
			ps.setInt(2, currentOrdinal)
			newOrdinal = ps.executeQuery { rs ->
				var last = currentOrdinal
				if (moveToStart) {
					while (rs.next()) {
						val current = rs.getInt(1)
						if (last - 1 > current)
							return@executeQuery last - 1
						last = current
					}
					if (last > 0)
						return@executeQuery 0
					else
						return@_moveUserInQueue Storage.Result_MoveToStartOrEnd.ON_EDGE
				} else {
					while (rs.next()) {
						val current = rs.getInt(1)
						if (last + 1 < current)
							return@executeQuery last + 1
						last = current
					}
					return@executeQuery last + 1
				}
			}
		}

		conn.prepareStatement(
			"""
				UPDATE "user->queue"
				SET ordinal = ?
				WHERE user = ? AND queue_set = ?
			"""
		) { ps ->
			ps.setInt(1, newOrdinal)
			ps.setLong(2, userDbId)
			ps.setLong(3, qs.dbId)
			ps.executeUpdate()
		}

		return Storage.Result_MoveToStartOrEnd.SUCCESSFUL(this._formatQueueSet(conn, qs))
	}

	override fun moveUserToStartOfQueue(chat: Long, message: Long, user: Long): Storage.Result_MoveToStartOrEnd =
		this._moveUserInQueue(chat, message, user, true)

	override fun moveUserToEndOfQueue(chat: Long, message: Long, user: Long): Storage.Result_MoveToStartOrEnd =
		this._moveUserInQueue(chat, message, user, false)


	override fun addModerator(chat: Long, user: Long): Boolean {
		TODO("Not yet implemented")
	}

	override fun removeModerator(chat: Long, message: Long): Boolean {
		TODO("Not yet implemented")
	}

	override fun getPermissions(chat: Long, user: Long): ModeratorPermissions? {
		TODO("Not yet implemented")
	}

	override fun setPermission(chat: Long, user: Long, permissionType: ModeratorPermissions.PermissionType, value: Boolean): Storage.Result_SetModeratorPermission {
		TODO("Not yet implemented")
	}
}